/*
 * Latency Tester: Use the TSC to measure variation in code execution caused by
 * hardware or the operating system
 *
 * (C) 2009 The Linux Foundation
 *      Christoph Lameter <cl@linux-foundation.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>
#include "ll.h"


int loop;
int histogram;
int verbose;
int runtime = 10;
int minimum = 1000;
int bucket = 100;

unsigned long samples;
unsigned long min, max, avg;
unsigned long zero_samples, beyond_max;
unsigned long scheduler_move;

#define MAX_US 10000

unsigned long hist[MAX_US];

struct rusage before, after;

void usage(void)
{
        printf("latencytest 4/17/2009 (C) 2009 Linux Foundation\n\n"
                "latencytest [-t<time>] [-b<bucketsize>][-d] [-h] [-l] [-v]\n"
                "-d|--diagram                      Show detailed distribution of latencies\n"
                "-l|--loop                         Repeat latency measurements at infinitum\n"
                "-h|--help                         Print usage information\n"
		"-t <n> | --time <n>               Maximum runtime (default 10 seconds)\n"
                "-v|--verbose                      Verbose output.\n"
		"-m <n> | --min <n>                Minimal value in Nanoseconds (default 1000).\n"
		"-b <n> | --bucket <n>             Bucket size in nanoseconds (default 100).\n"
        );
}

struct option opts[] = {
	{ "help", 0, NULL, 'h' },
	{ "verbose", 0, NULL, 'v' },
	{ "loop", 0, NULL, 'l' },
	{ "diagram", 0, NULL, 'd' },
	{ "time", 1, NULL, 't' },
	{ "min", 1, NULL, 'm' },
	{ "bucket", 1, NULL, 'b' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char *argv[])
{
	unsigned long long sum, sqsum;
	unsigned long long start, end, prior;
	unsigned long i;
	char c;
	int cpu = sched_getcpu();
	double avg;

	while ((c = getopt_long(argc, argv, "ldhvt:m:b:", opts, NULL)) != -1)
        switch (c) {
        case 'h' :
                usage();
                return 0;

        case 'd' :
                histogram = 1;
                break;

        case 'l' :
                loop = 1;
                break;

	case 'v':
		verbose = 1;
		break;
	case 't':
		runtime = atoi(optarg);
		break;

	case 'm':
		minimum = atoi(optarg);
		break;
	case 'b':
		bucket = atoi(optarg);
		break;

	default:
		fprintf(stderr, "%s: Invalid option '%c'\n", argv[0], optopt);
		return 1;
        }

        if (argc > optind) {
                usage();
                return 1;
        }

	ll_init(1);

redo:
	memset(hist, 0, sizeof(hist));
	zero_samples = 0;
	beyond_max = 0;
	min = max = avg = 0;
	sum = 0;
	sqsum = 0;

	if (getrusage(RUSAGE_SELF, &before)) {
		perror("getrusage");
		exit(1);
	}

	if (verbose)
		printf("Taking samples\n");

	start = timestamp_nanoseconds();
	end = start + runtime * 1000000000ULL;
	if (end < start) {
		fprintf(stderr, "Too long time period\n");
		abort();
	}

	prior = start;
	samples = 0;

	/* Obtain measurements */
	do {
		unsigned long long now = timestamp_nanoseconds();
		long diff = now - prior;

		if (cpu != sched_getcpu()) {
			if (verbose)
				printf("Scheduler moved latencytest from cpu %d to cpu %d\n",
					cpu, sched_getcpu());
			cpu = sched_getcpu();
			scheduler_move++;
			diff = 0;

		} else if (diff < 0) {
			printf("Negative! %ld (%lld - %lld)\n",
				 diff, now, prior);
			diff = 0;

		} else if (diff > (1UL << 31) ){
			printf("Untrackable outlier %.2f seconds (%lu) (%llu - %llu)\n",
				 diff / 1000000000.0, diff, now, prior);
			diff = 0;

		} else	if (diff < minimum)
			zero_samples++;
		else {
			if (diff < min || min == 0)
       	                	min = diff;
			if (diff > max)
                        	max = diff;
			if (diff / bucket > MAX_US)
				beyond_max++;
			else {
				hist [ diff / bucket] ++;
				sum += diff;
				sqsum += diff * diff;
				samples++;
			}
		}
		prior = now;
		cpu_relax();
	} while (prior < end);

	end = timestamp_nanoseconds();

	if (getrusage(RUSAGE_SELF, &after)) {
		perror("getrusage");
		exit(1);
	}

	if (verbose)
		printf("Finished sample taking\n");

	if (zero_samples)
		printf("%lu samples below %d nsec\n", zero_samples, minimum);

	if (beyond_max)
		printf("%lu samples above maximum for the histogram. Max is %d usecs\n",
			 beyond_max, MAX_US * bucket / 1000);

	if (scheduler_move)
		printf("%lu migrations to different processors\n", scheduler_move);

	if (before.ru_minflt != after.ru_minflt)
		printf("%ld minor faults\n", after.ru_minflt - before.ru_minflt);

	if (before.ru_majflt != after.ru_majflt)
		printf("%ld major faults\n", after.ru_majflt - before.ru_majflt);

	if (before.ru_nvcsw != after.ru_nvcsw)
		printf("%ld voluntary context switches\n", after.ru_nvcsw - before.ru_nvcsw);

	if (before.ru_nivcsw != after.ru_nivcsw)
		printf("%ld involuntary context switches\n", after.ru_nivcsw - before.ru_nivcsw);

	avg = (double)sum / samples;
	printf("%lu (%.8f%%) variances in %.2f seconds: minimum %.2fus maximum %.2fus average %.2fus stddev %.2fus\n",
			samples, samples * 100.0 / (zero_samples + samples),
			(end - start) / 1000000000.0,
                        min / 1000.0, max / 1000.0, avg / 1000.0,
			sqrt((double)sqsum / samples - avg * avg)/1000.0);

	if (histogram) {
		printf("usecs percent samples\n-------------------------\n");
		for (i = 0; i < MAX_US; i++)
			if (hist[i])
				printf("%4.1f %7.1f %8lu\n", i * (double)bucket / 1000.0,
					(hist[i] * 100.0) / samples,
					hist[i]);
	}

	if (loop)
		goto redo;
	return 0;
}

